# The file is part of the pyneovi project and is provided under the MIT License terms.
# For license information see LICENSE.txt.
"""
This module provides classes and functions to allow a network specification to
be created for use with pyneovi. A network specification defines the valid
networks for a given model of vehicle, along with the ECUs on those networks
and the identifiers and signals made available by the ECUs.

.. Author: John Kemp <kemp@kempj.co.uk>

Below is an example of a simple network specification. Only the medium speed
CAN bus is defined and only a single ECU is associated with that network. The
ECU provides four identifiers, each made up of a single signal.

.. literalinclude:: ../examples/create_network_spec.py

"""


import json


# Identifier signal types
ANALOG = 0

# Identifier value types
UNSIGNED_INT = 0

# Signal IO type
READ = 0
WRITE = 1
READ_WRITE = 2


class NonByteBoundarySignalError(Exception):
    pass
    

class ValueTooLargeError(Exception):
    pass


class UnsupportedValueTypeError(Exception):
    pass


def prepend_zeroes(values, final_length):
    return ([0] * (final_length - len(values))) + values


class _IdentifierList:
    def __init__(self, identifiers):
        self.by_name = {identifier.name: identifier for identifier in identifiers}
        self.by_id = {identifier.data_id: identifier for identifier in identifiers}
    
    def get_names(self):
        return self.by_name.keys()
    
    def get_ids(self):
        return self.by_id.keys()
    
    def __contains__(self, k):
        return k in self.by_name or k in self.by_id
    
    def __getitem__(self, k):
        if type(k) == int:
            return self.by_id[k]
        elif type(k) == str:
            return self.by_name[k]
        else:
            raise KeyError(k)


class Signal:
    """
    __init__(self, name='value', signal_type=ANALOG, value_type=UNSIGNED_INT, start=0, length=8, min=None, max=None, units='', m=1, b=0, description='')
    
    Represents a signal. An :py:class:`.spec.Identifier` will contain one or
    more signals.
    
    :param str name: The name of this signal.
    :param int signal_type: The type of signal. Currently, only
        :py:data:`.spec.ANALOG` is supported.
    :param int value_type: The type of value returned for the signal.
        Currently only :py:data:`.spec.UNSIGNED_INT` is supported.
    :param int start: The bit within the identifier's aggregate value at
        which this signal starts.
    :param int length: The length in bits of this signal.
    :param float min_val: The minimum value for this signal. This is not enforced
        and is for reference only. If not specified then b is used as the
        default value.
    :param float max_val:  The minimum value for this signal. This is not enforced
        and is for reference only. If not specified then (255. * m) + b is
        used as the default value.
    :param str units: The units of the signal once converted.
    :param float m: The slope for conversion of the signal into engineering
        units.
    :param float b: The offset for conversion of the signal into engineering
        units.
    :param str description: A human-readable description of the signal.
   
    :raises NonByteBoundarySignalError: If length or start are not integer
        multiples of 8.
    """
    def __init__(self, name='value', signal_type=ANALOG, value_type=UNSIGNED_INT, start=0, length=8, min_val=None, max_val=None, units='', m=1, b=0, description=''):
        self.name = name
        self.signal_type = signal_type
        self.value_type = value_type
        self.start = start
        self.length = length
        if min_val is None:
            self.min = b
        else:
            self.min = min_val
        if max_val is None:
            self.max = (255. * m) + b
        else:
            self.max = max_val
        self.units = units
        self.m = m
        self.b = b
        self.description = description
        
        if length % 8 != 0 or start % 8 != 0:
            raise NonByteBoundarySignalError
    
    def decode_signal_bytes(self, signal_bytes):
        """
        Converts the individual data bytes into the final value for the signal.
        
        :param signal_bytes: The bytes received from the ECU.
        :type signal_bytes: list of ints
        :returns: Tuple of final value and the units of the signal.
        
        :raises UnsupportedValueTypeError: If the signal is of a type that is
            currently unsupported.
        """
        if self.signal_type == ANALOG and self.value_type == UNSIGNED_INT:
            value = sum(x << (i*8) for i, x in enumerate(reversed(signal_bytes[self.start:self.start + (self.length / 8)])))
            value = self.m * value + self.b
        else:
            raise UnsupportedValueTypeError
        
        return value, self.units
    
    def encode_value(self, value):
        """
        Converts a value into the individual data bytes required for
        transmission.
        
        :param float/int value: The value for conversion.
        
        :raises UnsupportedValueTypeError: If the signal is of a type that is
            currently unsupported.
        :raises ValueTooLargeError: If the value does not fit into the number
            of bits specified for the signal.
        """
        if self.signal_type == ANALOG and self.value_type == UNSIGNED_INT:
            data_bytes = []
            value = int(round((value - self.b) / self.m))

            if value == 0:
                return prepend_zeroes([0], self.length / 8)
            
            while value > 0:
                data_bytes.append(value & 0xFF)
                value >>= 8
            
            if len(data_bytes) > (self.length / 8):
                raise ValueTooLargeError

            return prepend_zeroes([x for x in reversed(data_bytes)], self.length / 8)
        else:
            raise UnsupportedValueTypeError
    
    def __repr__(self):
        return 'Signal(name="%(name)s", signal_type=%(signal_type)u, value_type=%(value_type)u, start=%(start)u, length=%(length)u, min=%(min)s, max=%(max)s, units="%(units)s", m=%(m)f, b=%(b)f)' % self.__dict__


# TODO: Add method to encode all signal values into a final value to send
class Identifier:
    """
    __init__(self, name, data_id, signals = [], io_type = READ)
    
    Represents an identifier (an example of which would be "Blower Speed Output"
    in a HVAC ECU). The value of an identifier can consist of multiple
    individual signals but often there is only a single signal associated with
    an identifier.
    
    :param str name: The name of the identifier.
    :param int data_id: The address (internal to the ECU) with which this
        identifier is associated.
    :param signals: The signals that form this identifier.
    :type signals: list of spec.Signal
    :param int io_type: Allowed IO for this identifier: :py:data:`.spec.READ`,
        :py:data:`.spec.WRITE`, :py:data:`.spec.READ_WRITE`.
    """
    def __init__(self, name, data_id, signals=(), io_type=READ):
        self.name = name
        self.data_id = data_id
        self.data_id_bytes = [(data_id & 0xFF00) >> 8, data_id & 0x00FF]
        self.signals = signals
        self.io_type = io_type
    
    def get_data_id_bytes(self):
        """
        :returns: The data_id as individual bytes. Used when constructing a
            message.
        """
        return self.data_id_bytes
    
    def __repr__(self):
        return 'Identifier(name="%(name)s", data_id=%(data_id)X, signals=%(signals)s)' % self.__dict__
    
    def __str__(self):
        return 'Identifier: %(name)s (%(data_id)X)' % self.__dict__


class ECU:
    """
    Represents an ECU for the purpose of building a network specification. Note
    that if you need an object that allows you to communicate with the ECU then
    you're looking for :py:class:`.ecu.ECU`.
    
    :param int network: The network to transmit on. See NETID_* in
        :py:mod:`~neovi.neovi`.
    :param int address: The address of the ECU.
    :param str short_desc: A short name for the ECU (e.g. "HVAC").
    :param str long_desc: A long description for the ECU (e.g. including model
        number or other information).
    """
    def __init__(self, network, address, short_desc='', long_desc=''):
        self.network = network
        self.request_id = address
        self.response_id = address + 8
        self.short_desc = short_desc
        self.long_desc = long_desc
    
    def get_network(self):
        return self.network
    
    def get_request_id(self):
        return self.request_id
    
    def get_response_id(self):
        return self.response_id
    
    def __repr__(self):
        return 'ECU(network=%(network)u, address=0x%(request_id)X, short_desc="%(short_desc)s", long_desc="%(long_desc)s")' % self.__dict__
 
    
class PyNeoViJSONEncoder(json.JSONEncoder):
    """
    Can be passed to json.dump as the cls parameter to handle the additional
    object types defined here.
    """
    def default(self, o):
        if o.__class__.__name__ == 'Identifier':
            return 'PyNeoVi.Identifier', o.name, o.data_id, o.signals, o.io_type
        elif o.__class__.__name__ == 'ECU':
            return 'PyNeoVi.ECU', o.network, o.request_id, o.short_desc, o.long_desc
        elif o.__class__.__name__ == 'Signal':
            return 'PyNeoVi.Signal', o.name, o.signal_type, o.value_type, o.start, o.length, o.min, o.max, o.units, o.m, o.b, o.description
        else:
            return json.JSONEncoder.default(self, o)


def from_json(json_object):
    """
    Can be passed to json.load as the object_hook parameter to handle the
    additional object types defined here.
    """
    output = {}
    
    for k, v in json_object.items():
        if type(v) != list:
            if k == 'networks':
                output[k] = {int(k1): v1 for k1, v1 in v.items()}
            else:
                output[k] = v
        elif v[0] == 'PyNeoVi.ECU':
            output[k] = ECU(*v[1:])
        elif v[0] == 'PyNeoVi.Identifier':
            output[k] = Identifier(*v[1:])
        elif k == 'identifiers':
            new_idents = []
            for identifier in v:
                signals = [Signal(*signal_info[1:]) for signal_info in identifier[3]]
                new_idents.append(Identifier(identifier[1], identifier[2], signals))
            output[k] = _IdentifierList(new_idents)
        else:
            print('%s %s' % (k, v))
            output[k] = v
    
    return output




"""
    pyneovi is a Python wrapper around the API provided by Intrepid Control
    Systems for communicating with their NeoVI range of devices.

    Copyright © 2015-2018 John Kemp

    Web:   http://kempj.co.uk
    
    Email: kemp@kempj.co.uk

    pyneovi is provided under the MIT License terms. For license information see LICENSE.txt.
"""



# The file is part of the pyneovi project and is provided under the MIT License terms.
# For license information see LICENSE.txt.
"""
This module wraps the low-level interface to the neoVI range of devices. It is
unlikely that the functions defined here will be directly used as
:py:class:`.neodevice.NeoDevice` and related classes offer a more convenient
interface.

Of more interest will be the various constants defined here.

Device type identifiers (used by :py:func:`.neodevice.find_devices`):

======================= ====================== =====================
NEODEVICE_BLUE          NEODEVICE_SW_VCAN      NEODEVICE_FIRE
NEODEVICE_VCAN3         NEODEVICE_YELLOW       NEODEVICE_RED
NEODEVICE_ECU           NEODEVICE_IEVB         NEODEVICE_PENDANT
NEODEVICE_VIRTUAL_NEOVI NEODEVICE_ECUCHIP_UART NEODEVICE_PLASMA_1_11
NEODEVICE_FIRE_VNET     NEODEVICE_NEOANALOG    NEODEVICE_CT_OBD
NEODEVICE_PLASMA_1_12   NEODEVICE_PLASMA_1_13  NEODEVICE_ION_2
NEODEVICE_RADSTAR       NEODEVICE_ION_3        NEODEVICE_VCANFD
NEODEVICE_ECU15         NEODEVICE_ECU25        NEODEVICE_EEVB
NEODEVICE_VCANRF        NEODEVICE_FIRE2        NEODEVICE_FLEX
NEODEVICE_RADGALAXY     NEODEVICE_NEOECUCHIP
======================= ====================== =====================

Additional special identifiers are available:

==================== ================= =============
NEODEVICE_ANY_PLASMA NEODEVICE_ANY_ION NEODEVICE_ALL
==================== ================= =============

Network IDs:

=================================== =================================== ==========================
NETID_DEVICE

NETID_MSCAN                         NETID_SWCAN                         NETID_SWCAN2 
NETID_LSFTCAN                       NETID_LSFTCAN2                      NETID_HSCAN
NETID_HSCAN2                        NETID_HSCAN3                        NETID_HSCAN4
NETID_HSCAN5                        NETID_HSCAN6                        NETID_HSCAN7

NETID_LIN                           NETID_LIN2                          NETID_LIN3
NETID_LIN4                          NETID_LIN5                          NETID_LIN6

NETID_ETHERNET                      NETID_OP_ETHERNET1                  NETID_OP_ETHERNET2
NETID_OP_ETHERNET3                  NETID_OP_ETHERNET4                  NETID_OP_ETHERNET5
NETID_OP_ETHERNET6                  NETID_OP_ETHERNET7                  NETID_OP_ETHERNET8
NETID_OP_ETHERNET9                  NETID_OP_ETHERNET10                 NETID_OP_ETHERNET11
NETID_OP_ETHERNET12                 NETID_ETHERNET_DAQ

NETID_RS232                         NETID_UART                          NETID_UART2
NETID_UART3                         NETID_UART4

NETID_FLEXRAY                       NETID_FLEXRAY1A                     NETID_FLEXRAY1B
NETID_FLEXRAY2                      NETID_FLEXRAY2A                     NETID_FLEXRAY2B

NETID_3G_RESET_STATUS               NETID_3G_FB_STATUS                  NETID_3G_APP_SIGNAL_STATUS
NETID_3G_READ_DATALINK_CM_TX_MSG    NETID_3G_READ_DATALINK_CM_RX_MSG    NETID_3G_LOGGING_OVERFLOW
NETID_3G_READ_SETTINGS_EX

NETID_MOST                          NETID_MOST25                        NETID_MOST50
NETID_MOST150

NETID_FORDSCP                       NETID_J1708                         NETID_AUX
NETID_JVPW                          NETID_ISO                           NETID_ISO2
NETID_ISO3                          NETID_ISO4                          NETID_ISOPIC
NETID_MAIN51                        NETID_RED                           NETID_SCI
NETID_ISO14230                      NETID_RED_APP_ERROR                 NETID_CGI
NETID_DATA_TO_HOST                  NETID_TEXTAPI_TO_HOST               NETID_I2C1
NETID_SPI1                          NETID_RED_VBAT                      NETID_GMFSA
NETID_TCP
=================================== =================================== ==========================
"""


# http://www.intrepidcs.com/support/ICSDocumentation/neoVIDLL/neoVIDLLhelpdoc.html
# https://docs.python.org/2/library/ctypes.html#ctypes.c_int

# Python 2.x compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

from ctypes import *
import platform
import enum

import neovi.structures as structures


apilib = None


# noinspection PyUnusedLocal
def update_result(result, func, args):
    """
    Helper to convert an API call result into an error code.
    :param int result:
    :param func:
    :param tuple args:
    :return: Either SUCCESS or an error code.
    :rtype: int
    """
    if result == 0:
        return GetLastAPIError(args[0])[1]
    else:
        return SUCCESS


def _set_api_hints():
    apilib.icsneoFindNeoDevices.argtypes = [c_int, POINTER(structures.NeoDevice), POINTER(c_int)]
    apilib.icsneoOpenNeoDevice.argtypes = [POINTER(structures.NeoDevice), POINTER(c_int), POINTER(c_ubyte), c_int, c_int]
    apilib.icsneoClosePort.argtypes = [c_int, POINTER(c_int)]
    apilib.icsneoFreeObject.argtypes = [c_int]
    apilib.icsneoFreeObject.restype = None

    apilib.icsneoGetLastAPIError.argtypes = [c_int,  POINTER(c_int)]
    apilib.icsneoGetErrorMessages.argtypes = [c_int, POINTER(c_int), POINTER(c_int)]

    apilib.icsneoGetMessages.argtypes = [c_int, POINTER(structures.icsSpyMessage), POINTER(c_int), POINTER(c_int)]
    apilib.icsneoGetMessages.errcheck = update_result
    apilib.icsneoTxMessages.argtypes = [c_int, POINTER(structures.icsSpyMessage), c_int, c_int]
    apilib.icsneoTxMessages.errcheck = update_result
    apilib.icsneoWaitForRxMessagesWithTimeOut.argtypes = [c_int, c_uint]
    apilib.icsneoGetTimeStampForMsg.argtypes = [c_int, POINTER(structures.icsSpyMessage), POINTER(c_double)]
    apilib.icsneoGetTimeStampForMsg.errcheck = update_result
    apilib.icsneoEnableNetworkRXQueue.argtypes = [c_int, c_int]
    apilib.icsneoEnableNetworkRXQueue.errcheck = update_result
    apilib.icsneoGetISO15765Status.argtypes = [c_int, c_int, c_int, c_int, POINTER(c_int), POINTER(c_int)]
    apilib.icsneoGetISO15765Status.restype = None
    apilib.icsneoSetISO15765RxParameters.argtypes = [c_int, c_int, c_int, POINTER(structures.spyFilterLong), POINTER(structures.icsSpyMessage), c_int, c_int, c_int, c_int]
    apilib.icsneoSetISO15765RxParameters.restype = None

    apilib.icsneoGetConfiguration.argtypes = (c_int, POINTER(c_ubyte), POINTER(c_int))
    apilib.icsneoGetConfiguration.errcheck = update_result
    apilib.icsneoSendConfiguration.argtypes = (c_int, POINTER(c_ubyte), c_int)
    apilib.icsneoSendConfiguration.errcheck = update_result
    apilib.icsneoGetFireSettings.argtypes = (c_int, POINTER(structures.SFireSettings), c_int)
    apilib.icsneoGetFireSettings.errcheck = update_result
    apilib.icsneoSetFireSettings.argtypes = (c_int, POINTER(structures.SFireSettings), c_int, c_int)
    apilib.icsneoSetFireSettings.errcheck = update_result
    apilib.icsneoGetVCAN3Settings.argtypes = (c_int, POINTER(structures.SVCAN3Settings), c_int)
    apilib.icsneoGetVCAN3Settings.errcheck = update_result
    apilib.icsneoSetVCAN3Settings.argtypes = (c_int, POINTER(structures.SVCAN3Settings), c_int, c_int)
    apilib.icsneoSetVCAN3Settings.errcheck = update_result

    apilib.icsneoSetBitRate.argtypes = (c_int,  c_int, c_int)
    apilib.icsneoSetBitRate.errcheck = update_result

    apilib.icsneoGetHWFirmwareInfo.argtypes = (c_int, POINTER(structures.APIFirmwareInfo))
    apilib.icsneoGetDLLFirmwareInfo.argtypes = (c_int, POINTER(structures.APIFirmwareInfo))
    apilib.icsneoForceFirmwareUpdate.argtypes = (c_int,)
    apilib.icsneoForceFirmwareUpdate.restype = None


NEODEVICE_BLUE          = 0x00000001
NEODEVICE_SW_VCAN       = 0x00000002
NEODEVICE_DW_VCAN       = 0x00000004
NEODEVICE_FIRE          = 0x00000008
NEODEVICE_VCAN3         = 0x00000010
NEODEVICE_YELLOW        = 0x00000020
NEODEVICE_RED           = 0x00000040
NEODEVICE_ECU           = 0x00000080
NEODEVICE_IEVB          = 0x00000100
NEODEVICE_PENDANT       = 0x00000200
NEODEVICE_VIRTUAL_NEOVI = 0x00000400
NEODEVICE_ECUCHIP_UART  = 0x00000800
NEODEVICE_PLASMA_1_11   = 0x00001000
NEODEVICE_FIRE_VNET     = 0x00002000
NEODEVICE_NEOANALOG     = 0x00004000
NEODEVICE_CT_OBD        = 0x00008000
NEODEVICE_PLASMA_1_12   = 0x00010000
NEODEVICE_PLASMA_1_13   = 0x00020000
NEODEVICE_ION_2         = 0x00040000
NEODEVICE_RADSTAR       = 0x00080000
NEODEVICE_ION_3         = 0x00100000
NEODEVICE_VCANFD        = 0x00200000
NEODEVICE_ECU15         = 0x00400000
NEODEVICE_ECU25         = 0x00800000
NEODEVICE_EEVB          = 0x01000000
NEODEVICE_VCANRF        = 0x02000000
NEODEVICE_FIRE2         = 0x04000000
NEODEVICE_FLEX          = 0x08000000
NEODEVICE_RADGALAXY     = 0x10000000

NEODEVICE_ANY_PLASMA    = NEODEVICE_PLASMA_1_11 | NEODEVICE_FIRE_VNET | NEODEVICE_PLASMA_1_12 | NEODEVICE_PLASMA_1_13
NEODEVICE_ANY_ION       = NEODEVICE_ION_2 | NEODEVICE_ION_3
NEODEVICE_ALL           = 0xFFFFBFFF

NEODEVICE_NEOECUCHIP    = NEODEVICE_IEVB


# For general error checks (all values >= 0 are defined error codes)
SUCCESS = -1

# For wait_for_rx_messages_with_timeout (all values >= 0 are defined error codes)
NO_MESSAGES = -2
MESSAGES_RECVD = -1


NETID_DEVICE                        = 0
NETID_HSCAN                         = 1
NETID_MSCAN                         = 2
NETID_SWCAN                         = 3
NETID_LSFTCAN                       = 4
NETID_FORDSCP                       = 5
NETID_J1708                         = 6
NETID_AUX                           = 7
NETID_JVPW                          = 8
NETID_ISO                           = 9
NETID_ISOPIC                        = 10
NETID_MAIN51                        = 11
NETID_RED                           = 12
NETID_SCI                           = 13
NETID_ISO2                          = 14
NETID_ISO14230                      = 15
NETID_LIN                           = 16
NETID_OP_ETHERNET1                  = 17
NETID_OP_ETHERNET2                  = 18
NETID_OP_ETHERNET3                  = 19
NETID_ISO3                          = 41
NETID_HSCAN2                        = 42
NETID_HSCAN3                        = 44
NETID_OP_ETHERNET4                  = 45
NETID_OP_ETHERNET5                  = 46
NETID_ISO4                          = 47
NETID_LIN2                          = 48
NETID_LIN3                          = 49
NETID_LIN4                          = 50
NETID_MOST                          = 51
NETID_RED_APP_ERROR                 = 52
NETID_CGI                           = 53
NETID_3G_RESET_STATUS               = 54
NETID_3G_FB_STATUS                  = 55
NETID_3G_APP_SIGNAL_STATUS          = 56
NETID_3G_READ_DATALINK_CM_TX_MSG    = 57
NETID_3G_READ_DATALINK_CM_RX_MSG    = 58
NETID_3G_LOGGING_OVERFLOW           = 59
NETID_3G_READ_SETTINGS_EX           = 60
NETID_HSCAN4                        = 61
NETID_HSCAN5                        = 62
NETID_RS232                         = 63
NETID_UART                          = 64
NETID_UART2                         = 65
NETID_UART3                         = 66
NETID_UART4                         = 67
NETID_SWCAN2                        = 68
NETID_ETHERNET_DAQ                  = 69
NETID_DATA_TO_HOST                  = 70
NETID_TEXTAPI_TO_HOST               = 71
NETID_I2C1                          = 71
NETID_SPI1                          = 72
NETID_OP_ETHERNET6                  = 73
NETID_RED_VBAT                      = 74
NETID_OP_ETHERNET7                  = 75
NETID_OP_ETHERNET8                  = 76
NETID_OP_ETHERNET9                  = 77
NETID_OP_ETHERNET10                 = 78
NETID_OP_ETHERNET11                 = 79
NETID_FLEXRAY1A                     = 80
NETID_FLEXRAY1B                     = 81
NETID_FLEXRAY2A                     = 82
NETID_FLEXRAY2B                     = 83
NETID_LIN5                          = 84
NETID_FLEXRAY                       = 85
NETID_FLEXRAY2                      = 86
NETID_OP_ETHERNET12                 = 87
NETID_MOST25                        = 90
NETID_MOST50                        = 91
NETID_MOST150                       = 92
NETID_ETHERNET                      = 93
NETID_GMFSA                         = 94
NETID_TCP                           = 95
NETID_HSCAN6                        = 96
NETID_HSCAN7                        = 97
NETID_LIN6                          = 98
NETID_LSFTCAN2                      = 99


class BitRate(enum.IntEnum):
    """
    Available bit rates for NeoVI network interfaces.
    """
    BR_20000 = 0
    BR_33333 = 1
    BR_50000 = 2
    BR_62500 = 3
    BR_83333 = 4
    BR_100000 = 5
    BR_125000 = 6
    BR_250000 = 7
    BR_500000 = 8
    BR_800000 = 9
    BR_1000000 = 10


DEFAULT_BIT_RATE = BitRate.BR_500000


def network_name(nid):
    """
    Lookup the friendly name for a network ID.
    :param int nid: The network ID to look up.
    :return: The friendly name for the network.
    :rtype: str
    """
    networks = {
        1: 'HSCAN',
        2: 'MSCAN',
        3: 'SWCAN',
        4: 'LSFTCAN'
    }
    return networks.get(nid, 'Unknown')


def InitializeAPI():
    """
    Initialise the API and prepares it for use. It must be called before any
    other API calls are made.
    
    Note that the underlying neoVI library only requires this on Linux, but
    pyneovi requires it on all platforms. This provides a consistent interface
    without platform-specific checks in application code and also allows
    pyneovi to perform its own initialisation.
    """
    global apilib
    
    if platform.system() == 'Linux':
        apilib = CDLL('libicsneoAPI.so.0.1.3')
    else:
        apilib = windll.icsneo40
    
    _set_api_hints()

    if platform.system() == 'Linux':
        apilib.icsneoInitializeAPI()


def ShutdownAPI():
    """
    Shut down the API and releases any resources allocated during its use.
    
    Note that the underlying neoVI library only requires this on Linux, but
    pyneovi requires it on all platforms. This provides a consistent interface
    without platform-specific checks in application code.
    """
    if platform.system() == 'Linux':
        apilib.icsneoShutdownAPI()
    
    
def GetLastAPIError(hObject):
    """
    Get the error generated by the last API call.
    
    API errors are generated and stored on a 'per-thread' basis. The calling
    thread will only receive errors generated within it's own context. If a new
    API error is generated before the previous error has been retrieved, the
    previous error will be lost. All errors generated can still be retrieved
    using GetErrorMessages. However, GetErrorMessages will return errors
    generated in all threads, not just the current thread.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
        
    :return: A 2-tuple consisting of i) a boolean indicating if a new error was
        generated since the port was opened or GetLastAPIError was last called
        and ii) the error generated by the last API call.
    """
    error_num = c_int()

    result = apilib.icsneoGetLastAPIError(hObject,  byref(error_num))
    return bool(result), error_num.value
        
        
def GetErrorMessages(hObject):
    """
    Read the neoVI DLL error message queue. The error queue will be reset after
    this function is called.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
        
    :return: A 2-tuple consisting of i) a boolean indicating success or failure
        of the function and ii) a list of errors generated by all threads.
    """
    msgs = (c_int * 600)()
    num_errors = c_int()

    result = apilib.icsneoGetErrorMessages(hObject, msgs, byref(num_errors))
    return bool(result), msgs[:num_errors.value]
        
        
def FindNeoDevices(types):
    """
    Find any attached neoVI devices.
    
    :param int types: Filter the results to given types of devices. The
        NEODEVICE_* constants can be OR'd together.
        
    :return: A list of :py:class:`.structures.NeoDevice` objects.
    """
    num_devices = c_int(255)
    devices = (structures.NeoDevice * 255)()

    result = apilib.icsneoFindNeoDevices(types, devices, byref(num_devices))
    
    if result == 0:
        return []

    return devices[:num_devices.value]


def OpenNeoDevice(device, network_ids=None, config_read=True):
    """
    Open a communication link the a neoVI device.
    
    :param structures.NeoDevice device: A device information structure returned
        by :py:func:`.neovi.FindNeoDevices`.
    :param network_ids: An array of network IDs to assign to the available
        networks. The default is to use the predefined network IDs.
    :param bool config_read: Should the device's configuration be read before
        enabling it? It is recommended that this be set to true.
        
    :return: An object handle for use in other API functions or None if the
        call failed.
    """
    handle = c_int()

    result = apilib.icsneoOpenNeoDevice(device, byref(handle), network_ids, c_int(config_read), 0)
    
    if result == 1:
        return handle
    else:
        return None
        

def ClosePort(hObject):
    """
    Close the communication link with the neoVI hardware.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    
    :return: A 2-tuple containing i) a bool representing the success or failure
        of the function and ii) a list of error codes if the bool is false
        (empty otherwise). See
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation.
    """
    num_errors = c_int()
    
    result = apilib.icsneoClosePort(hObject, byref(num_errors))
    if num_errors > 0:
        return bool(result), GetErrorMessages(hObject)[1]
    else:
        return bool(result), []
        
        
def FreeObject(hObject):
    """
    Release any resources that were allocated by OpenNeoDevice.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    """
    apilib.icsneoFreeObject(hObject)


def GetMessages(hObject):
    """
    Read messages from the neoVI device. The driver object will hold 20,000
    received messages before it will generate an RX buffer overflow error.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    
    :return: A 3-tuple containing i) either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation), ii) a list of
        :py:class:`.structures.icsSpyMessage` objects, and iii) a list of error
        codes obtained via :py:func:`.neovi.GetErrorMessages`.
    """
    msgs = (structures.icsSpyMessage * 20000)()
    num_msgs = c_int()
    num_errors = c_int()
    
    result = apilib.icsneoGetMessages(hObject, msgs, byref(num_msgs), byref(num_errors))
    
    if num_errors > 0:
        errors = GetErrorMessages(hObject)[1]
    else:
        errors = []
    
    return result, msgs[:num_msgs.value], errors
    

def TxMessages(hObject, msgs, network, num_msgs=1):
    """
    Transmit messages asynchronously to vehicle networks using the neoVI
    hardware.
    
    After the messages has been transmitted there will be a transmit report
    message returned from the device. The transmit report will be read out with
    :py:func:`.neovi.GetMessages`. Any message read which has the
    SPY_STATUS_TX_MSG bit set in the status bitfield is a transmit report. 

    You can also identify a particular transmitted message with DescriptionID
    field. This two byte field (only 14 bits are used) allows the programmer to
    assign an arbitrary number to a message. This number is then returned in
    the transmit report.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    :param msgs: A list of :py:class:`.structures.icsSpyMessage` objects.
    :param int network: The network to transmit the message on.
    :param int num_msgs: The number of messages to transmit. This should always
        be 1 unless you are transmitting a long Message on ISO or J1708.
    
    :return: Either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation).
    """
    return apilib.icsneoTxMessages(hObject, msgs, network, num_msgs)
    
    
def WaitForRxMessagesWithTimeOut(hObject, timeout_ms):
    """
    Wait a specified amount of time for received messages.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    :param int timeout_ms: A timeout in ms to wait for a received message
        before returning.
    
    :return: Either MESSAGES_RECVD, NO_MESSAGES, or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation).
    """
    result = apilib.icsneoWaitForRxMessagesWithTimeOut(hObject, timeout_ms)
    
    if result == -1:
        return GetLastAPIError(hObject)[1]
    elif result == 0:
        return NO_MESSAGES
    else:
        return MESSAGES_RECVD


def GetTimeStampForMsg(hObject, msg):
    """
    Calculate the timestamp for a message, based on the connected hardware type, and convert it to a usable variable.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    :param msg: A :py:class:`.structures.icsSpyMessage` object.
    
    :return: A 2-tuple of i) either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation) and ii) the calculated timestamp.
    """
    timestamp = c_double()
    
    result = apilib.icsneoGetTimeStampForMsg(hObject, msg, byref(timestamp))
    
    return result, timestamp.value
    
    
def EnableNetworkRXQueue(hObject, enable):
    """
    Enable or disable the received message queue (and thus reception of
    network traffic). This applies only to the application that called this
    function, other applications are not affected.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    :param bool enable: Specifies whether to enable (true) or disable (false)
        reception of network traffic.
        
    :return: Either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation).
    """
    return apilib.icsneoEnableNetworkRXQueue(hObject, c_int(enable))
    
    
def GetISO15765Status(hObject, network, clear_rx_status):
    """
    Get, and optionally clear, the current receive status of the CAN ISO15765-2
    network layer.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    :param int network: Which CAN network the status is requested for.
    :param bool clear_rx_status: If true will clear the receive status and
        reset the receive state machine.
        
    :return: The receive status. See
        `GetISO15765Status <https://cdn.intrepidcs.net/support/neoVIDLL/GetISO15765Status.htm>`_
        in the ICS API documentation.
    """
    reserved = c_int()
    rx_status = c_int()
    
    apilib.icsneoGetISO15765Status(hObject, network, 0, c_int(clear_rx_status), byref(reserved), byref(rx_status))
    return rx_status


def SetISO15765RxParameters(hObject, network, enable, msg_filter, flow_ctx_msg, timeout_ms, flow_control_block_size, uses_extended_addressing):
    """
    set the parameters necessary to control CAN ISO15765-2 network layer message reception.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    :param int network: The network to transmit the message on.
    :param bool enable: If true ISO15765 services will be enabled for this network.
    
    See
    `SetISO15765RxParameters <https://cdn.intrepidcs.net/support/neoVIDLL/SetISO15765RxParameters.htm>`_
    in the ICS API documentation for full descriptions of the remaining parameters.
    """
    apilib.icsneoSetISO15765RxParameters(hObject, network, c_int(enable), msg_filter, flow_ctx_msg, timeout_ms, flow_control_block_size, uses_extended_addressing, 0)


def GetConfiguration(hObject):
    """
    Read the configuration from a NeoVI Blue or ValueCAN device.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    
    :return: A 2-tuple of i) either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation) and ii) an array of 1024 configuration
        bytes (see
        `Configuration Array <https://cdn.intrepidcs.net/support/neoVIDLL/ConfigArray.htm>`_
        in the ICS API documentation).
    """
    pData = (c_ubyte * 1024)()
    plNumBytes = c_int()

    result = apilib.icsneoGetConfiguration(hObject, pData, plNumBytes)
    
    return result, pData[:plNumBytes.value]


def SendConfiguration(hObject, pData):
    """
    Send configuration information to a NeoVI Blue or ValueCAN device.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    :param pData: An array of configuration bytes (see
        `Configuration Array <https://cdn.intrepidcs.net/support/neoVIDLL/ConfigArray.htm>`_
        in the ICS API documentation).
    
    :return: Either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation).
    """
    result = apilib.icsneoSendConfiguration(hObject, pData, sizeof(pData))
    
    return result
    

def GetFireSettings(hObject):
    """
    Read the configuration settings from a neoVI Fire device.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    
    :return: A 2-tuple of i) either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation) and ii) a
        :py:class:`.structures.SFireSettings` object.
    """
    pSettings = structures.SFireSettings()

    result = apilib.icsneoGetFireSettings(hObject, pSettings, sizeof(pSettings))
    
    return result, pSettings
    

def SetFireSettings(hObject, pSettings, bSaveToEEPROM):
    """
    Write configuration settings to a neoVI Fire device.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    :param structures.SFireSettings pSettings: The configuration to write.
    :param bool bSaveToEEPROM: Overwrite the stored EEPROM values so that these
        settings persist across power-cycles.
    
    :return: Either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation).
    """
    result = apilib.icsneoSetFireSettings(hObject, pSettings, sizeof(pSettings), c_int(bSaveToEEPROM))
    
    return result


def GetVCAN3Settings(hObject):
    """
    Read the configuration settings from a ValueCAN3 device.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    
    :return: A 2-tuple of i) either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation) and ii) a
        :py:class:`.structures.SVCAN3Settings` object.
    """
    pSettings = structures.SVCAN3Settings()

    result = apilib.icsneoGetVCAN3Settings(hObject, pSettings, sizeof(pSettings))
    
    return result, pSettings


def SetVCAN3Settings(hObject, pSettings, bSaveToEEPROM):
    """
    Write configuration settings to a ValueCAN3 device.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    :param structures.SVCAN3Settings pSettings: The configuration to write.
    :param bool bSaveToEEPROM: Overwrite the stored EEPROM values so that these
        settings persist across power-cycles.
    
    :return: Either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation).
    """
    result = apilib.icsneoSetVCAN3Settings(hObject, pSettings, sizeof(pSettings), c_int(bSaveToEEPROM))
    
    return result


def SetBitRate(hObject, iBitRate, iNetworkID):
    """
    Set the bit rates for networks.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    :param int iBitRate: The new bit rate setting.
    :param int iNetworkID: The network to set the bit rate on.
    
    :return: Either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation).
    """
    result = apilib.icsneoSetBitRate(hObject, iBitRate, iNetworkID)
    
    return result
    
    
def GetHWFirmwareInfo(hObject):
    """
    Return the firmware version of the open neoVI device.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    
    :return: A 2-tuple of i) either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation) and ii) a
        :py:class:`.structures.APIFirmwareInfo` object.
    """
    pInfo = structures.APIFirmwareInfo()
    result = apilib.icsneoGetHWFirmwareInfo(hObject, pInfo)
    return bool(result), pInfo
    
    
def GetDLLFirmwareInfo(hObject):
    """
    Return the firmware version stored within the DLL API.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    
    :return: A 2-tuple of i) either SUCCESS or an error code (see
        `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
        in the ICS API documentation) and ii) a
        :py:class:`.structures.APIFirmwareInfo` object.
    """
    pInfo = structures.APIFirmwareInfo()
    result = apilib.icsneoGetDLLFirmwareInfo(hObject, pInfo)
    return bool(result), pInfo
    
    
def ForceFirmwareUpdate(hObject):
    """
    Force the firmware on a neoVI device to be updated to the version stored in the DLL API.
    
    :param hObject: An object handle returned by
        :py:func:`.neovi.OpenNeoDevice`.
    """
    apilib.icsneoForceFirmwareUpdate(hObject)
    

def GetDLLVersion():
    """
    Return the software version of the DLL.
    
    :return: The version of the API DLL as an integer.
    :rtype: int
    """
    return apilib.icsneoGetDLLVersion()




# The file is part of the pyneovi project and is provided under the MIT License terms.
# For license information see LICENSE.txt.
"""
.. literalinclude:: ../neovi/can.py
"""

import enum


# Message types
DIAG_CURRENT_DATA          = 0x01
DIAG_FREEZE_FRAME_DATA     = 0x02
DIAG_STORED_DTCS           = 0x03
DIAG_CLEAR_DTCS            = 0x04
DIAG_PENDING_DTCS          = 0x07
DIAG_VEHICLE_INFORMATION   = 0x09
DIAG_PERMANENT_DTCS        = 0x0A

DIAGNOSTIC_SESSION_CONTROL = 0x10
ECU_RESET                  = 0x11
SECURITY_ACCESS            = 0x27
COMMUNICATION_CONTROL      = 0x28
TESTER_PRESENT             = 0x3E
ACCESS_TIMING_PARAMETER    = 0x83
SECURED_DATA_TRANSMISSION  = 0x84
CONTROL_DTC_SETTING        = 0x85
RESPONSE_ON_EVENT          = 0x86
LINK_CONTROL               = 0x87

READ_DATA_BY_ID            = 0x22
READ_MEMORY_BY_ADDRESS     = 0x23
READ_SCALING_DATA_BY_ID    = 0x24
READ_DATA_BY_PERIODIC_ID   = 0x2A
DYNAMICALLY_DEFINE_DATA_ID = 0x2C
WRITE_DATA_BY_ID           = 0x2E
WRITE_MEMORY_BY_ADDRESS    = 0x3D

CLEAR_DIAGNOSTIC_INFORMATION = 0x14
READ_DTC_INFORMATION         = 0x19

IO_CONTROL_BY_ID           = 0x2F

ROUTINE_CONTROL            = 0x31

REQUEST_DOWNLOAD           = 0x34
REQUEST_UPLOAD             = 0x35
TRANSFER_DATA              = 0x36
REQUEST_TRANSFER_EXIT      = 0x37


# DIAG_VEHICLE_INFORMATION PIDs
SUPPORTED_PIDS                 = 0x00
VIN_MESSAGE_COUNT              = 0x01
VIN                            = 0x02
CALIBRATION_ID_COUNT           = 0x03
CALIBRATION_ID                 = 0x04
CALIBRATION_VERIFICATION_COUNT = 0x05
CALIBRATION_VERIFICATION       = 0x06
PERFORMANCE_TRACKING_COUNT_S   = 0x07
PERFORMANCE_TRACKING           = 0x08
ECU_NAME_COUNT                 = 0x09
ECU_NAME                       = 0x0A
PERFORMANCE_TRACKING_COUNT_C   = 0x0B


class SessionType(enum.IntEnum):
    """
    Session types that can be requested with a DIAGNOSTIC_SESSION_CONTROL message.
    """
    Default                = 0x01
    Programming            = 0x02
    ExtendedDiagnostic     = 0x03
    SafetySystemDiagnostic = 0x04


class ControlType(enum.IntEnum):
    """
    Control types that can be used with an ID_CONTROL_BY_ID message.
    """
    ReturnControlToECU  = 0x00
    ResetToDefault      = 0x01
    FreezeCurrentState  = 0x02
    ShortTermAdjustment = 0x03


error_code_lookup = {
    0x10: 'General reject',
    0x11: 'Service not supported',
    0x12: 'Subfunction not supported',
    0x13: 'Invalid message length or format',
    0x14: 'Response too long',
    0x21: 'Busy repeat request',
    0x22: 'System state not correct (is ignition on?)',
    0x24: 'Request sequence error',
    0x25: 'No response from sub-net component',
    0x26: 'Failure prevented execution',
    0x31: 'Bad message format / request out of range',
    0x33: 'Security access denied',
    0x35: 'Invalid key',
    0x36: 'Exceeded number of attempts',
    0x37: 'Required time delay not expired',
    0x70: 'Upload/download not accepted',
    0x71: 'Transfer data suspended',
    0x72: 'General programming failure',
    0x73: 'Wrong Block Sequence Counter',
    0x78: 'Request ok, response pending',
    0x7E: 'Sub-function not supported in active session',
    0x7F: 'Service not supported in active session',
    0x80: 'Not supported in active session',   # Not standard, where did this come from?
}




# The file is part of the pyneovi project and is provided under the MIT License terms.
# For license information see LICENSE.txt.
"""
Provides the :py:class:`.ECU` class that represents an individual ECU.

.. Author: John Kemp <kemp@kempj.co.uk>

Below is an example of reading values from an ECU. Note that it requires a
network spec such as the one created in the example on the :py:mod:`.spec` module page.

.. literalinclude:: ../examples/read_values.py
"""


# Python 2.x compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import threading
import logging

from neovi.structures import format_array, get_status_bits_set, array_equal, icsSpyMessage
import neovi.can as can


class CommandError(Exception):
    """
    An error response was received after sending a message on the bus.
    """
    def __init__(self, code):
        self.code = code
        if code in can.error_code_lookup:
            self.desc = can.error_code_lookup[code]
            message = '0x%X (%s)' % (self.code, can.error_code_lookup[self.code])
        else:
            self.desc = ''
            message = '0x%X' % self.code

        super(CommandError, self).__init__(message)


class NoResponseError(Exception):
    """
    No response was received after sending a message on the bus.
    """
    pass


class ECU:
    """
    Represents a vehicle ECU.
    
    :param spec.ECU ecu_info: Details of the ECU (network, address...).
    :param neodevice.NeoDevice interface: The neoVI device to use for
        communication.
    :param bool auto_tester_present_on_auth: Should a "tester present" message
        be sent at a regular interval if successful authentication (by an
        external tool) is detected?
    :param bool log_unhandled_messages: Should messages to this ECU that have
        no associated subscription be logged?
    """
    def __init__(self, ecu_info, interface, auto_tester_present_on_auth=True, log_unhandled_messages=False):
        self.ecu = ecu_info['ecu']
        self.identifiers = ecu_info['identifiers']
        self.interface = interface

        self.tester_present_interval = None
        self.tester_present_timer = None
        self.tester_present_lock = threading.Lock()

        self.waiting_for = []
        self.waiting_for_lock = threading.Lock()
        self.log_unhandled_messages = log_unhandled_messages

        self.msg_lock = threading.Lock()

        self.auto_tester_present_on_auth = auto_tester_present_on_auth

        self.interface.subscribe_to(self._handle_incoming_message, self.ecu.get_network(), self.ecu.get_response_id())

    def _get_identifier_bytes(self, identifier):
        if identifier in self.identifiers:
            return self.identifiers[identifier].get_data_id_bytes()
        elif identifier.__class__.__name__ == 'Identifier':
            return identifier.get_data_id_bytes()
        else:
            return identifier

    def _get_identifier_object(self, identifier):
        if identifier in self.identifiers:
            return self.identifiers[identifier]
        elif identifier.__class__.__name__ == 'Identifier':
            return identifier
        else:
            return None

    # noinspection PyMethodMayBeStatic
    def _msg_recv_callback(self, msg, misc):
        misc[1][0] = msg
        misc[0].set()

    def _transaction(self, msg_type, match_bytes, send_func, send_params):
        event = threading.Event()
        msg = [None]
        subs_key = (msg_type + 0x40, match_bytes, self._msg_recv_callback, (event, msg))

        self.msg_lock.acquire()

        self.waiting_for_lock.acquire()
        self.waiting_for.append(subs_key)
        self.waiting_for_lock.release()

        send_func(*send_params)
        result = event.wait(0.2)

        if not result:
            # Give up waiting for the message and clean up our request
            self.waiting_for_lock.acquire()
            self.waiting_for.remove(subs_key)
            self.waiting_for_lock.release()
            self.msg_lock.release()
            raise NoResponseError

        self.msg_lock.release()

        # msg (as passed around in subs_key) should have been modified by the thread handling the incoming messages.
        msg = msg[0]    # type: icsSpyMessage

        if msg.Data[1] == 0x7F:
            raise CommandError(msg.Data[3])

        return msg

    def _send_read_data_by_id(self, identifier):
        """
        Send a "Read Data By ID" message.
        
        :param identifier: An array of integers specifying the identifier to
                           read (e.g. [0x98, 0x05]).
        """
        return self.interface.tx_message(self.ecu.get_network(), self.ecu.get_request_id(), can.READ_DATA_BY_ID, identifier)

    def read_data_by_id(self, identifier):
        """
        Send a "Read Data By ID" message and return the response value(s). The
        return value will be a dictionary of decoded values if an identifier
        name or an identifier object is passed in, or the raw message bytes if
        an integer array is passed in.
        
        :param identifier: This may be one of three types of object/value:
           1) the name of an identifier known to the :py:class:`.spec.ECU` that
           was passed to the constructor, 
           2) a :py:class:`.spec.Identifier` object, or 
           3) an array of integers specifying the identifier to read (e.g.
           [0x98, 0x05]).
        """
        id_bytes = self._get_identifier_bytes(identifier)

        msg = self._transaction(can.READ_DATA_BY_ID, id_bytes, self._send_read_data_by_id, (id_bytes,))

        ident_obj = self._get_identifier_object(identifier)
        if ident_obj is None:
            return msg.Data[4:]

        values = {}
        for signal in ident_obj.signals:
            values[signal.name] = signal.decode_signal_bytes(msg.Data[4:])

        return values

    def _send_io_control_by_id(self, identifier, value=None, control_type=can.ControlType.ShortTermAdjustment):
        """
        Send a "IO Control By ID" message.
        
        :param identifier: An array of integers specifying the identifier to read (e.g. [0x98, 0x05]).
        :param value: The value to set the output to. A value of None will use the default. Default = [0x00].
        :param can.ControlType control_type: The type of control to apply. Default = ControlType.ShortTermAdjustment.
        """
        if control_type == can.ControlType.ShortTermAdjustment:
            if value is None:
                value = [0x00]
            return self.interface.tx_message(self.ecu.get_network(), self.ecu.get_request_id(), can.IO_CONTROL_BY_ID, identifier + [control_type] + value)
        else:
            return self.interface.tx_message(self.ecu.get_network(), self.ecu.get_request_id(), can.IO_CONTROL_BY_ID, identifier + [control_type])

    def io_control_by_id(self, identifier, value=0x00, control_type=can.ControlType.ShortTermAdjustment):
        """
        Send a "IO Control By ID" message.
        
        :param identifier: This may be one of three types of object/value:
           1) the name of an identifier known to the :py:class:`.spec.ECU` that was passed to the constructor,
           2) a :py:class:`.spec.Identifier` object, or
           3) an array of integers specifying the identifier to read (e.g. [0x98, 0x05]).
           Note that due to limitations of the current implementation, the value will be assumed to correspond to the
           first signal defined for the identifier.
        :param int value: The value to set the output to. Conversion of the value into the form required by the
                          underlying signal will be performed here. Default = 0x00.
        :param can.ControlType control_type: The type of control to apply. Default = SHORT_TERM_ADJUSTMENT.
        """
        id_bytes = self._get_identifier_bytes(identifier)

        if control_type == can.ControlType.ShortTermAdjustment:
            value = self._get_identifier_object(identifier).signals[0].encode_value(value)

        msg = self._transaction(can.IO_CONTROL_BY_ID, id_bytes, self._send_io_control_by_id, (id_bytes, value, control_type))

        # Figure out what we want to return here
        return True

    def _send_diagnostic_session_control(self, session_type):
        """
        Send a "Diagnostic Session Control" message.
        
        :param can.SessionType session_type: The type of session to switch to.
        """
        return self.interface.tx_message(self.ecu.get_network(), self.ecu.get_request_id(), can.DIAGNOSTIC_SESSION_CONTROL, [session_type])

    def diagnostic_session_control(self, session_type):
        """
        Send a "Diagnostic Session Control" message.

        :param can.SessionType session_type: The type of session to switch to.
        """
        msg = self._transaction(can.DIAGNOSTIC_SESSION_CONTROL, [], self._send_diagnostic_session_control, (session_type,))
        return list(msg.Data[3:])

    def _send_security_access(self, subfunction, key):
        return self.interface.tx_message(self.ecu.get_network(), self.ecu.get_request_id(), can.SECURITY_ACCESS, [subfunction] + key)

    def security_access(self, subfunction, key=None):
        if key is None:
            key = []
        msg = self._transaction(can.SECURITY_ACCESS, [subfunction], self._send_security_access, (subfunction, key))

        if subfunction % 2 == 0:
            # We were sending a key
            return True
        else:
            # We were requesting a seed
            seed_length = msg.Data[0] - 2
            return list(msg.Data[3:3 + seed_length])

    def send_tester_present(self):
        """
        Send a "Tester Present" message to avoid a diagnostic session timing out.
        """
        self.interface.tx_message(self.ecu.get_network(), self.ecu.get_request_id(), can.TESTER_PRESENT, [0x80])

    def send_periodic_tester_present(self, interval=2):
        """
        Start sending a periodic "tester present" message to avoid a diagnostic session timing out.
        :param int interval: Period in seconds between messages.
        """
        self.tester_present_lock.acquire()

        # To avoid all sorts of nonsense, we just don't do anything if the
        # timer is already set up. Call stop_periodic_tester_present before
        # calling this again.
        if self.tester_present_timer is not None:
            self.tester_present_lock.release()
            return

        self.send_tester_present()

        self.tester_present_interval = interval
        self.tester_present_timer = threading.Timer(interval, self._send_periodic_tester_present)
        self.tester_present_timer.start()

        self.tester_present_lock.release()

    def _send_periodic_tester_present(self):
        self.tester_present_lock.acquire()
        # We do this check to avoid the following:
        #  - stop_periodic_tester_present is entered and obtains the lock.
        #  - The timer fires, enters this method, and blocks on the lock.
        #  - stop_periodic_tester_present cancels the timer, and releases the lock.
        #  - This method continues and spawns another timer, ensuring that the signal continues.
        # We, in fact, *can* stop the signal.
        if self.tester_present_interval is not None:
            self.send_tester_present()
            self.tester_present_timer = threading.Timer(self.tester_present_interval, self._send_periodic_tester_present)
            self.tester_present_timer.start()
        self.tester_present_lock.release()

    def stop_periodic_tester_present(self):
        """
        Stop sending the periodic "tester present" message.
        """
        self.tester_present_lock.acquire()

        if self.tester_present_timer is not None:
            self.tester_present_timer.cancel()
            self.tester_present_timer = None

        self.tester_present_interval = None

        self.tester_present_lock.release()

    def _handle_incoming_message(self, msg, user_data):
        handled = False
        self.waiting_for_lock.acquire()
        # First, check if it's an error message, if so then we have to search differently
        if msg.Data[1] == 0x7F:
            for i, (msg_type, additional_bytes, callback, misc) in enumerate(self.waiting_for):
                if msg_type == (msg.Data[2] + 0x40):
                    callback(msg, misc)
                    self.waiting_for.pop(i)
                    handled = True
                    break
        else:
            for i, (msg_type, additional_bytes, callback, misc) in enumerate(self.waiting_for):
                if msg_type == msg.Data[1] and array_equal(msg.Data[2:2 + len(additional_bytes)], additional_bytes):
                    callback(msg, misc)
                    self.waiting_for.pop(i)
                    handled = True
                    break
        self.waiting_for_lock.release()

        # Special handling for tester present
        if self.auto_tester_present_on_auth and [x for x in msg.Data][:3] == [0x02, 0x67, 0x04]:
            print("Successful auth detected, starting periodic tester present signal")
            self.send_periodic_tester_present()

        if self.log_unhandled_messages and not handled:
            logging.debug('Unhandled message from ECU %s' % self.ecu.short_desc)
            logging.debug(' - ArbIDOrHeader = %X, number data bytes = %X' % (msg.ArbIDOrHeader, msg.NumberBytesData))
            logging.debug(' - Data = %s' % format_array(msg.Data))
            logging.debug(' - Ack = %s' % format_array(msg.AckBytes))
            logging.debug(' - Value = %f, misc data = %X' % (msg.Value, msg.MiscData))
            stat, stat2 = get_status_bits_set(msg)
            stat = stat + stat2
            for i in range(0, len(stat), 3):
                stats = stat[i:i+3]
                logging.debug(' - %s' % ', '.join(stats))

    def __del__(self):
        self.stop_periodic_tester_present()




# The file is part of the pyneovi project and is provided under the MIT License terms.
# For license information see LICENSE.txt.
"""
This module provides classes representing neoVI devices, all derived from core
functionality in the :py:class:`.NeoDevice` class.

.. Author: John Kemp <kemp@kempj.co.uk>

The following neoVI classes are defined:

.. inheritance-diagram::
      NeoDevice
      NeoFire
      NeoRed
      NeoVCAN3
      NeoBlue
      NeoDWVCAN
      
Below is a simple example that logs all messages to and from a specified ECU.

.. literalinclude:: ../examples/log_all.py
"""


# Python 2.x compatibility
from __future__ import absolute_import, division, print_function, unicode_literals

import threading
from ctypes import sizeof

import neovi.neovi as neovi
import neovi.structures as structures


def init_api():
    """
    Load the neoVI API library and initialise it. This must be called before
    making any API calls.
    """
    neovi.InitializeAPI()
    

def find_devices(types=neovi.NEODEVICE_ALL, auto_open=False):
    """
    find_devices(types=neovi.NEODEVICE_ALL, auto_open=False)

    Find any attached neoVI devices.
    
    :param int types: Filter the results to given types of devices. The
        NEODEVICE_* constants defined in :py:mod:`~neovi.neovi` can be OR'd together.
    :param bool auto_open: Determines whether each device should be
        automatically opened. If this is False then you must call
        :py:meth:`.NeoDevice.open` in order to open a given device.
    """
    return [device_class_lookup[device.DeviceType](device, auto_open) for device in neovi.FindNeoDevices(types)]
    

class OpenFailedError(Exception):
    """
    Failed to open a NeoVI device.
    """
    pass


class InvalidConfigurationError(Exception):
    """
    An invalid configuration was supplied to be written to the device.
    """
    pass


class NeoDevice:
    """
    Represents a generic neoVI device, providing an interface for transmitting
    and receiving messages as well as configuring the device.
    
    :param structures.NeoDevice device: Device identifier, as returned by
        :py:func:`.neovi.FindNeoDevices` for example. See also
        :py:func:`.find_devices` which returns a list of pre-constructed
        NeoDevice objects.
    :param bool auto_open: Determines whether the device should be
        automatically opened. If this is False then you must call
        :py:meth:`.NeoDevice.open` in order to open the device.
    :param bool launch_msg_queue_thread: Determines whether to start a
        thread to receive incoming messages. If this is False then you must
        process the message queue via the
        :py:meth:`.NeoDevice._process_msg_queue` method or fetch the
        messages via the :py:meth:`.NeoDevice.get_messages` method.
    """
    def __init__(self, device, auto_open=False, launch_msg_queue_thread=True):
        self._device = device
        self._handle = None
        self._tx_id = 1
        self._subscriptions = []
        self._subscriptions_all = []
        self._msg_queue_thread = None
        self._pump_messages = False
        self._subscriptions_lock = threading.Lock()
        self._subscriptions_all_lock = threading.Lock()
        
        if auto_open:
            self.open(launch_msg_queue_thread)
            
    def open(self, launch_msg_queue_thread=True):
        """
        Open the device and optionally launch the message thread. This is only
        required if the object was constructed with auto_open = False.
        
        :param bool launch_msg_queue_thread: Determines whether to start a
            thread to receive incoming messages. If this is False then you must
            process the message queue via the
            :py:meth:`.NeoDevice._process_msg_queue` method or fetch the
            messages via the :py:meth:`.NeoDevice.get_messages` method.
        
        :raises OpenFailedError: If the device cannot be opened.
        """
        self._handle = neovi.OpenNeoDevice(self._device)
        if self._handle is None:
            raise OpenFailedError
        if launch_msg_queue_thread:
            self._start_msg_queue_thread()
    
    def get_type(self):
        """
        :return: The type of device represented. See NEODEVICE_* in
            :py:mod:`~neovi.neovi`.
        """
        return self._device.DeviceType
    
    def get_settings(self):
        """
        Get the current device configuration.
        :return:
        """
        raise NotImplementedError
        
    def set_settings(self, settings, save_to_eeprom=False):
        """
        Set the device configuration.
        :param settings: A device-specific configuration.
        :param save_to_eeprom: Determines if the configuration will be persisted across device power cycles.
        :return:
        """
        raise NotImplementedError
        
    def tx_raw_message(self, msg, network_id):
        """
        Transmits a pre-constructed message.
        
        :param icsSpyMessage msg: The message to transmit.
        :param int network_id: The network to transmit on. See NETID_* in
            :py:mod:`~neovi.neovi`.
        
        :return: Status code. :py:data:`.neovi.SUCCESS` or an error code. See
            `TxMessages <https://cdn.intrepidcs.net/support/neoVIDLL/TxMessages.htm>`_
            in the ICS API documentation
            for possible codes (constants not yet defined within pyneovi).
        """
        return neovi.TxMessages(self._handle, msg, network_id)
    
    def tx_message(self, network_id, dest, msg_type, payload):
        """
        Transmits a pre-constructed message.
        
        :param int network_id: The network to transmit on. See NETID_* in
            :py:mod:`~neovi.neovi`.
        :param int dest: The address of the destination ECU.
        :param int msg_type: The message type to send. See :py:mod:`.can`.
        :param payload: Up to 6 bytes to send.
        :type payload: list of ints
        
        :return: Tuple of status code and transmission id. Status code is
            :py:data:`.neovi.SUCCESS` or an error code. See
            `TxMessages <https://cdn.intrepidcs.net/support/neoVIDLL/TxMessages.htm>`_
            in the ICS API documentation
            for possible codes (constants not yet defined within pyneovi).
            Transmission id can be safely ignored.
        """
        msg = structures.icsSpyMessage()
        msg.ArbIDOrHeader = dest
        msg.NumberBytesData = 8
        for i in range(8):
            msg.Data[i] = 0
        msg.Data[0] = len(payload) + 1
        msg.Data[1] = msg_type
        for i in range(len(payload)):
            msg.Data[i + 2] = payload[i]
        msg.StatusBitField = 0
        msg.StatusBitField2 = 0
        msg.DescriptionID = self._tx_id
        self._tx_id += 1
        return self.tx_raw_message(msg, network_id), self._tx_id - 1
        
    def subscribe_to(self, callback, network=None, address=None, msg_type=None, additional_bytes=None, auto_remove=False, user_data=None):
        """
        Set a callback function to be called upon reception of a defined subset
        of messages. Note that this will only occur if the message thread has
        been started or if :py:meth:`.NeoDevice._process_msg_queue` is called
        manually. All parameters other than callback, auto_remove, and user_data
        define filtering criteria that will be used to determine whether to
        pass the message to this callback.
        
        :param func callback: The method or function to be called. This must
            take two parameters: 1) a message as a
            :py:class:`.structures.icsSpyMessage` object, and 2) a user data
            parameter with no pre-defined meaning within pyneovi.
        :param int network: The network ID of interest (see NETID_* in
            :py:mod:`~neovi.neovi`). If None (the default) then it will be
            ignored for filtering.
        :param int address: The address of the ECU of interest. If None (the
            default) then it will be ignored for filtering.
        :param int msg_type: The message type of interest (see :py:mod:`.can`).
            If None (the default) then it will be ignored for filtering.
        :param additional_bytes: Additional payload bytes to use for filtering.
            May be an empty list. A value of None (the default) will be converted
            to an empty list automatically.
        :type additional_bytes: list of ints
        :param bool auto_remove: If True then the subscription is removed once
            the first message matching the provided criteria has been received.
        :param user_data: This parameter has no pre-defined meaning within
            pyneovi - the value is passed to the callback function along with
            the received message.
        """
        self._subscriptions_lock.acquire()
        self._subscriptions.append((callback, network, address, msg_type, [] if additional_bytes is None else additional_bytes, auto_remove, user_data))
        self._subscriptions_lock.release()
    
    def subscribe_to_all(self, callback):
        """
        Set a callback function to be called upon reception of any message. Note
        that this will only occur if the message thread has been started or if
        :py:meth:`.NeoDevice._process_msg_queue` is called manually.
        
        :param func callback: The method or function to be called. This must
            take three parameters: 1) a statue code of either
            :py:data:`.neovi.SUCCESS` or an error code (see
            `TxMessages <https://cdn.intrepidcs.net/support/neoVIDLL/TxMessages.htm>`_
            in the ICS API documentation
            for possible codes (constants not yet defined within pyneovi)), 2)
            a list of messages as :py:class:`.structures.icsSpyMessage` objects,
            and 3) a list of errors as integers (see
            `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
            in the ICS API documentation
            for a complete error list).
        """
        self._subscriptions_all_lock.acquire()
        self._subscriptions_all.append(callback)
        self._subscriptions_all_lock.release()
    
    def get_messages(self):
        """
        Fetch pending received messages. Note that if the message thread was
        launched then this should not be called.
        
        :returns: A tuple containing 1) a statue code of either
            :py:data:`.neovi.SUCCESS` or an error code (see
            `TxMessages <https://cdn.intrepidcs.net/support/neoVIDLL/TxMessages.htm>`_
            in the ICS API documentation
            for possible codes (constants not yet defined within pyneovi)), 2)
            a list of messages as :py:class:`.structures.icsSpyMessage` objects,
            and 3) a list of errors as integers (see
            `Error Messages <https://cdn.intrepidcs.net/support/neoVIDLL/apiErrorMessages.htm>`_
            in the ICS API documentation
            for a complete error list).
        """
        return neovi.GetMessages(self._handle)
    
    def setup_networks(self, network_settings):
        # We don't know how to handle anything other than the Fire right now.
        # Also, this should probably move into the subclasses.
        if self.get_type() != neovi.NEODEVICE_FIRE:
            return
        
        settings = self.get_settings()
        
        if neovi.NETID_HSCAN in network_settings:
            settings.can1.Mode = 0
            settings.can1.SetBaudrate = 0
            settings.can1.Baudrate = network_settings[neovi.NETID_HSCAN]['bitrate']
        else:
            settings.can1.Mode = 1
        
        if neovi.NETID_MSCAN in network_settings:
            settings.can2.Mode = 0
            settings.can2.SetBaudrate = 0
            settings.can2.Baudrate = network_settings[neovi.NETID_MSCAN]['bitrate']
        else:
            settings.can2.Mode = 1
        
        self.set_settings(settings)
    
    def get_firmware_version(self):
        result, info = neovi.GetHWFirmwareInfo(self._handle)
        return info if result else None

    def get_dll_firmware_version(self):
        result, info = neovi.GetDLLFirmwareInfo(self._handle)
        return info if result else None

    def force_firmware_update(self):
        neovi.ForceFirmwareUpdate(self._handle)

    def close(self):
        """
        Close the device and shutdown the message thread (if there is one).
        """
        self._pump_messages = False
        if self._msg_queue_thread is not None:
            self._msg_queue_thread.join()

        if self._handle is not None:
            neovi.ClosePort(self._handle)
            neovi.FreeObject(self._handle)
            self._handle = None

    def _process_msg_queue(self):
        result, msgs, errors = self.get_messages()

        self._subscriptions_all_lock.acquire()
        for callback in self._subscriptions_all:
            callback(result, msgs, errors)
        self._subscriptions_all_lock.release()

        self._subscriptions_lock.acquire()
        for msg in msgs:
            for i in range(len(self._subscriptions) - 1, -1, -1):
                callback, network, address, msg_type, additional_bytes, auto_remove, user_data = self._subscriptions[i]
                if ((network is None or network == msg.NetworkID) and
                        (address is None or address == msg.ArbIDOrHeader) and
                        (msg_type is None or msg_type == msg.Data[1]) and
                        (additional_bytes is None or structures.array_equal(additional_bytes, msg.Data[2:2 + len(additional_bytes)]))):
                    callback(msg, user_data)
                    if auto_remove:
                        del self._subscriptions[i]
        self._subscriptions_lock.release()

    def _start_msg_queue_thread(self):
        self._pump_messages = True
        self._msg_queue_thread = threading.Thread(target=self._msg_queue_thread_func)
        self._msg_queue_thread.start()

    def _msg_queue_thread_func(self):
        while self._pump_messages:
            self._process_msg_queue()
    
    def __del__(self):
        self.close()
            

class NeoFire(NeoDevice):
    """
    Represents a neoVI Fire device. Should be used if settings specific to the
    neoVI Fire must be read/written.
    """
    def __init__(self, device, auto_open=True, launch_msg_queue_thread=True):
        NeoDevice.__init__(self, device, auto_open, launch_msg_queue_thread)
       
    def get_settings(self):
        return neovi.GetFireSettings(self._handle)
    
    def set_settings(self, settings, save_to_eeprom=False):
        return neovi.SetFireSettings(self._handle, settings, save_to_eeprom)


class NeoRed(NeoFire):
    """
    Represents a neoVI Red device. Should be used if settings specific to the
    neoVI Red must be read/written.
    """
    pass


class NeoVCAN3(NeoDevice):
    """
    Represents a ValueCAN3 device. Should be used if settings specific to the
    ValueCAN3 must be read/written.
    """
    def __init__(self, device, auto_open=True, launch_msg_queue_thread=True):
        NeoDevice.__init__(self, device, auto_open, launch_msg_queue_thread)
       
    def get_settings(self):
        return neovi.GetVCAN3Settings(self._handle)
    
    def set_settings(self, settings, save_to_eeprom=False):
        return neovi.SetVCAN3Settings(self._handle, settings, save_to_eeprom)
        
        
class NeoBlue(NeoDevice):
    """
    Represents a neoVI Blue device. Should be used if settings specific to the
    neoVI Blue must be read/written.
    """
    def __init__(self, device, auto_open=True, launch_msg_queue_thread=True):
        NeoDevice.__init__(self, device, auto_open, launch_msg_queue_thread)
    
    def set_settings(self, settings, save_to_eeprom=False):
        if sizeof(settings) != 1024:
            raise InvalidConfigurationError
        return neovi.SendConfiguration(self._handle, settings)


class NeoDWVCAN(NeoBlue):
    pass


# noinspection PyPep8
device_class_lookup = {
    neovi.NEODEVICE_FIRE    : NeoFire,
    neovi.NEODEVICE_RED     : NeoRed,
    neovi.NEODEVICE_VCAN3   : NeoVCAN3,
    neovi.NEODEVICE_BLUE    : NeoBlue,
    neovi.NEODEVICE_DW_VCAN : NeoDWVCAN
}




from setuptools import setup

readme_file = open('README.txt', 'r')
long_description = '\n'.join(readme_file.readlines())
readme_file.close()

setup(name='pyneovi',
      packages=['neovi'],
      version='1.0',
      description='A wrapper around the API provided by Intrepid Control Systems for communicating with their NeoVI range of devices',
      long_description=long_description,
      author='John Kemp',
      author_email='kemp@kempj.co.uk',
      url='http://kempj.co.uk/projects/pyneovi/',
      license='MIT',
      classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
      ]
)



# The file is part of the pyneovi project and is provided under the MIT License terms.
# For license information see LICENSE.txt.
"""
Classes provided by this module:
 * :class:`NeoDevice`
 * :class:`icsSpyMessage`
 * :class:`spyFilterLong`
 * :class:`CAN_SETTINGS`
 * :class:`SWCAN_SETTINGS`
 * :class:`LIN_SETTINGS`
 * :class:`ISO9141_KEYWORD2000__INIT_STEP`
 * :class:`ISO9141_KEYWORD2000_SETTINGS`
 * :class:`UART_SETTINGS`
 * :class:`STextAPISettings`
 * :class:`SFireSettings`
 * :class:`SVCAN3Settings`

.. Author: John Kemp <kemp@kempj.co.uk>
"""


from ctypes import *


def array_equal(a1, a2):
    if len(a1) != len(a2):
        print("Mismatched array sizes!")
        return False
    for i1, i2 in zip(a1, a2):
        if i1 != i2:
            return False
    return True


def format_array(array):
    return '%X %X %X %X %X %X %X %X' % (array[0], array[1], array[2], array[3], array[4], array[5], array[6], array[7])
    

def get_status_bits_set(msg):
    values = [
        (0x01, 'GLOBAL_ERR'),
        (0x02, 'TX_MSG'),
        (0x04, 'XTD_FRAME'),
        (0x08, 'REMOTE_FRAME'),
        (0x10, 'CRC_ERROR'),
        (0x20, 'CAN_ERROR_PASSIVE'),
        (0x40, 'INCOMPLETE_FRAME'),
        (0x80, 'LOST_ARBITRATION'),
        (0x100, 'UNDEFINED_ERROR'),
        (0x200, 'CAN_BUS_OFF'),
        (0x400, 'CAN_ERROR_WARNING'),
        (0x800, 'BUS_SHORTED_PLUS'),
        (0x1000, 'BUS_SHORTED_GND'),
        (0x2000, 'CHECKSUM_ERROR'),
        (0x4000, 'BAD_MESSAGE_BIT_TIME_ERROR'),
        (0x8000, 'IFR_DATA'),
        (0x10000, 'HARDWARE_COMM_ERROR'),
        (0x20000, 'EXPECTED_LEN_ERROR'),
        (0x40000, 'INCOMING_NO_MATCH'),
        (0x80000, 'BREAK'),
        (0x100000, 'AVSI_REC_OVERFLOW'),
        (0x200000, 'TEST_TRIGGER'),
        (0X400000, 'AUDIO_COMMENT'),
        (0x800000, 'GPS_DATA'),
        (0x1000000, 'ANALOG_DIGITAL_INPUT'),
        (0x2000000, 'TEXT_COMMENT'),
        (0x4000000, 'NETWORK_MESSAGE_TYPE'),
        (0x8000000, 'VSI_TX_UNDERRUN'),
        (0x10000000, 'VSI_IFR_CRC_Bit'),
        (0x20000000, 'INIT_MESSAGE'),
        (0x40000000, 'HIGH_SPEED_MESSAGE')
    ]
    
    values2 = [
        (0x01, 'HAS_VALUE'),
        (0x02, 'VALUE_IS_BOOLEAN'),
        (0x04, 'HIGH_VOLTAGE'),
        (0x08, 'LONG_MESSAGE')
    ]
    
    bits_set = []
    bits_set2 = []
    
    for v in values:
        if msg.StatusBitField & v[0] == v[0]:
            bits_set.append(v[1])
    
    for v in values2:
        if msg.StatusBitField2 & v[0] == v[0]:
            bits_set2.append(v[1])
    
    return bits_set, bits_set2


class NeoDevice(Structure):
    _fields_ = [('DeviceType',        c_ulong),
                ('Handle',            c_int),
                ('NumberOfClients',   c_int),
                ('SerialNumber',      c_int),
                ('MaxAllowedClients', c_int)]
                

class icsSpyMessage(Structure):
    _fields_ = [('StatusBitField',      c_ulong),
                ('StatusBitField2',     c_ulong),
                ('TimeHardware',        c_ulong),
                ('TimeHardware2',       c_ulong),
                ('TimeSystem',          c_ulong),
                ('TimeSystem2',         c_ulong),
                ('TimeStampHardwareID', c_ubyte),
                ('TimeStampSystemID',   c_ubyte),
                ('NetworkID',           c_ubyte),
                ('NodeID',              c_ubyte),
                ('Protocol',            c_ubyte),
                ('MessagePieceID',      c_ubyte),
                ('ColorID',             c_ubyte),
                ('NumberBytesHeader',   c_ubyte),
                ('NumberBytesData',     c_ubyte),
                ('DescriptionID',       c_short),
                ('ArbIDOrHeader',       c_long),
                ('Data',                c_ubyte * 8),
                ('AckBytes',            c_ubyte * 8),
                ('Value',               c_float),
                ('MiscData',            c_ubyte)]
    
    def __repr__(self):
        fields = ['StatusBitField', 'StatusBitField2', 'NodeID', 'DescriptionID', 'ArbIDOrHeader', 'Value']
        field_values = ['%s=%s' % (k, v) for k, v in [(row, getattr(self, row)) for row in fields]]
        data_values = ['%s=(%X %X %X %X %X %X %X %X)' % (k, v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7]) for k, v in [(row, getattr(self, row)) for row in ['Data', 'AckBytes']]]
        return 'icsSpyMessage(%s)' % ', '.join(field_values + data_values)
                

class spyFilterLong(Structure):
    _fields_ = [('StatusValue',          c_ulong),
                ('StatusMask',           c_ulong),
                ('Status2Value',         c_ulong),
                ('Status2Mask',          c_ulong),
                ('Header',               c_ulong),
                ('HeaderMask',           c_ulong),
                ('MiscData',             c_ulong),
                ('MiscDataMask',         c_ulong),
                ('ByteDataMSB',          c_ulong),
                ('ByteDataLSB',          c_ulong),
                ('ByteDataMaskMSB',      c_ulong),
                ('ByteDataMaskLSB',      c_ulong),
                ('HeaderLength',         c_ulong),
                ('ByteDataLength',       c_ulong),
                ('NetworkID',            c_ulong),
                ('FrameMaster',          c_ushort),
                ('bUseArbIdRangeFilter', c_ubyte),
                ('bStuff2',              c_ubyte),
                ('ExpectedLength',       c_ulong),
                ('NodeID',               c_ulong)]
                
                
class CAN_SETTINGS(Structure):
    _fields_ = [('Mode',        c_ubyte),
                ('SetBaudrate', c_ubyte),
                ('Baudrate',    c_ubyte),
                ('NetworkType', c_ubyte),
                ('TqSeg1',      c_ubyte),
                ('TqSeg2',      c_ubyte),
                ('TqProp',      c_ubyte),
                ('TqSync',      c_ubyte),
                ('BRP',         c_ushort),
                ('auto_baud',   c_ushort)]


class SWCAN_SETTINGS(Structure):
    _fields_ = [('Mode',                   c_ubyte),
                ('SetBaudrate',            c_ubyte),
                ('Baudrate',               c_ubyte),
                ('NetworkType',            c_ubyte),
                ('TqSeg1',                 c_ubyte),
                ('TqSeg2',                 c_ubyte),
                ('TqProp',                 c_ubyte),
                ('TqSync',                 c_ubyte),
                ('BRP',                    c_ushort),
                ('high_speed_auto_switch', c_ushort),
                ('auto_baud',              c_ushort)]


class LIN_SETTINGS(Structure):
    _fields_ = [('Baudrate',       c_uint),
                ('spbrg',          c_ushort),
                ('brgh',           c_ushort),
                ('MasterResistor', c_ubyte),
                ('Mode',           c_ubyte)]


class ISO9141_KEYWORD2000__INIT_STEP(Structure):
    _fields_ = [('time_500us', c_ushort),
                ('k',          c_ushort),
                ('l',          c_ushort)]


class ISO9141_KEYWORD2000_SETTINGS(Structure):
    _fields_ = [('Baudrate',        c_uint),
                ('spbrg',           c_ushort),
                ('brgh',            c_ushort),
                ('init_steps',      ISO9141_KEYWORD2000__INIT_STEP * 16),
                ('init_step_count', c_ubyte),
                ('p2_500us',        c_ushort),
                ('p3_500us',        c_ushort),
                ('p4_500us',        c_ushort),
                ('chksum_enabled',  c_ushort)]


class UART_SETTINGS(Structure):
    _fields_ = [('Baudrate',     c_ushort),
                ('spbrg',        c_ushort),
                ('brgh',         c_ushort),
                ('parity',       c_ushort),
                ('stop_bits',    c_ushort),
                ('flow_control', c_ubyte),  # 0- off, 1 - Simple CTS RTS,
                ('reserved_1',   c_ubyte),
                ('bOptions',     c_uint)]   # AND to combind these values invert_tx = 1 invert_rx = 2 half_duplex = 4


class STextAPISettings(Structure):
    _fields_ = [('can1_tx_id',      c_uint),
                ('can1_rx_id',      c_uint),
                ('can1_options',    c_uint),    # Set to 1 for Extended, 0 for standard
                ('can2_tx_id',      c_uint),
                ('can2_rx_id',      c_uint),
                ('can2_options',    c_uint),    # Set to 1 for Extended, 0 for standard
                ('network_enables', c_uint),
                ('can3_tx_id3',     c_uint),
                ('can3_rx_id3',     c_uint),
                ('can3_options',    c_uint),    # Set to 1 for Extended, 0 for standard
                ('can4_tx_id4',     c_uint),
                ('can4_rx_id4',     c_uint),
                ('can4_options',    c_uint),    # Set to 1 for Extended, 0 for standard
                ('Reserved0',       c_int),
                ('Reserved1',       c_int),
                ('Reserved2',       c_int),
                ('Reserved3',       c_int),
                ('Reserved4',       c_int)]


class SFireSettings(Structure):
    _fields_ = [('can1',                            CAN_SETTINGS),
                ('can2',                            CAN_SETTINGS),
                ('can3',                            CAN_SETTINGS),
                ('can4',                            CAN_SETTINGS),
                ('swcan',                           SWCAN_SETTINGS),
                ('lsftcan',                         CAN_SETTINGS),
                ('lin1',                            LIN_SETTINGS),
                ('lin2',                            LIN_SETTINGS),
                ('lin3',                            LIN_SETTINGS),
                ('lin4',                            LIN_SETTINGS),
                ('cgi_enable_reserved',             c_ushort),
                ('cgi_baud',                        c_ushort),
                ('cgi_tx_ifs_bit_times',            c_ushort),
                ('cgi_rx_ifs_bit_times',            c_ushort),
                ('cgi_chksum_enable',               c_ushort),
                ('network_enables',                 c_ushort),
                ('network_enabled_on_boot',         c_ushort),
                ('pwm_man_timeout',                 c_ushort),
                ('pwr_man_enable',                  c_ushort),
                ('misc_io_initial_ddr',             c_ushort),
                ('misc_io_initial_latch',           c_ushort),
                ('misc_io_analog_enable',           c_ushort),
                ('misc_io_report_period',           c_ushort),
                ('misc_io_on_report_events',        c_ushort),
                ('ain_sample_period',               c_ushort),
                ('ain_threshold',                   c_ushort),
                ('iso15765_separation_time_offset', c_ushort),
                ('iso9141_kwp_enable_reserved',     c_ushort),
                ('iso9141_kwp_settings',            ISO9141_KEYWORD2000_SETTINGS),
                ('perf_en',                         c_ushort),
                ('iso_parity',                      c_ushort),
                ('iso_msg_termination',             c_ushort),
                ('iso_tester_pullup_enable',        c_ushort),
                ('network_enables_2',               c_ushort),
                ('iso9141_kwp_settings2',           ISO9141_KEYWORD2000_SETTINGS),
                ('iso_parity_2',                    c_ushort), 
                ('iso_msg_termination_2',           c_ushort),
                ('iso9141_kwp_settings_3',          ISO9141_KEYWORD2000_SETTINGS),
                ('iso_parity_3',                    c_ushort), 
                ('iso_msg_termination_3',           c_ushort), 
                ('iso9141_kwp_settings_4',          ISO9141_KEYWORD2000_SETTINGS),
                ('iso_parity_4',                    c_ushort), 
                ('iso_msg_termination_4',           c_ushort), 
                ('fast_init_network_enables_1',     c_ushort),
                ('fast_init_network_enables_2',     c_ushort),
                ('uart',                            UART_SETTINGS),
                ('uart2',                           UART_SETTINGS),
                ('text_api',                        STextAPISettings)]
                
                
class SVCAN3Settings(Structure):
    _fields_ = [('can1',                            CAN_SETTINGS),
                ('can2',                            CAN_SETTINGS),
                ('network_enables',                 c_ushort),
                ('network_enabled_on_boot',         c_ushort),
                ('iso15765_separation_time_offset', c_short),
                ('perf_en',                         c_ushort),
                ('misc_io_initial_ddr',             c_ushort),
                ('misc_io_initial_latch',           c_ushort),
                ('misc_io_report_period',           c_ushort),
                ('misc_io_on_report_events',        c_ushort)]


class APIFirmwareInfo(Structure):
    _fields_ = [('iType',                           c_int),
                # Date and Time (2nd generation neoVI only) 
                ('iMainFirmDateDay',                c_int),
                ('iMainFirmDateMonth',              c_int),
                ('iMainFirmDateYear',               c_int),
                ('iMainFirmDateHour',               c_int),
                ('iMainFirmDateMin',                c_int),
                ('iMainFirmDateSecond',             c_int),
                ('iMainFirmChkSum',                 c_int),
    
                # Version data (3rd generation neoVI only) 
                ('iAppMajor',                       c_ubyte),
                ('iAppMinor',                       c_ubyte),
                ('iManufactureDay',                 c_ubyte),
                ('iManufactureMonth',               c_ubyte),
                ('iManufactureYear',                c_ushort),
                ('iBoardRevMajor',                  c_ubyte),
                ('iBoardRevMinor',                  c_ubyte),
                ('iBootLoaderVersionMajor',         c_ubyte),
                ('iBootLoaderVersionMinor',         c_ubyte)]
